from matplotlib import pyplot as plt
import numpy as np

from outset._auxlib.calc_aspect_ import calc_aspect
from outset._auxlib.set_aspect_ import set_aspect


def test_set_aspect_square():
    # note: asserts inside of set_aspect will run
    plt.gca().set_ylim(0, 1)
    plt.gca().set_xlim(0, 1)
    set_aspect(plt.gca(), 0.5) == 1

    plt.gca().set_xlim(0, 1)
    plt.gca().set_ylim(1, 2)
    set_aspect(plt.gca(), 2) == 1


def test_set_aspect_tall():
    # note: asserts inside of set_aspect will run
    plt.gca().set_xlim(0, 1)
    plt.gca().set_ylim(0, 2)
    set_aspect(plt.gca(), 1)

    plt.gca().set_xlim(0, 1)
    plt.gca().set_ylim(0, 2)
    set_aspect(plt.gca(), 4)

    plt.gca().set_xlim(0, 1)
    plt.gca().set_ylim(0, 2)
    set_aspect(plt.gca(), 0.5)


def test_set_aspect_wide():
    # note: asserts inside of set_aspect will run
    plt.gca().set_ylim(0, 1)
    plt.gca().set_xlim(0, 2)
    set_aspect(plt.gca(), 1)

    plt.gca().set_ylim(0, 1)
    plt.gca().set_xlim(0, 2)
    set_aspect(plt.gca(), 0.25)

    plt.gca().set_ylim(0, 1)
    plt.gca().set_xlim(0, 2)
    set_aspect(plt.gca(), 2)



import dataclasses
import typing


@dataclasses.dataclass
class SplitKwarg:
    """Wraps a kwarg value that should take on distinct values for 'source' and 'outset' axes when maping or broadcasting over an `outset.OutsetGrid`.

    Attributes
    ----------
    source : typing.Any
        The value of the argument to be used in the 'source' context.
    outEDset : typing.Any
        The value of the argument to be used in the 'outset' context.
    """

    source: typing.Any
    outset: typing.Any



import typing

import numpy as np


class CalcBoundsIQR:
    """Functor to set outlier bounds a fixed ratio above/below interquartile
    range.

    Bounds are calculated as (`q1` - `iqr_multiplier` * `iqr`, `q3 +
    iqr_multiplier` * `iqr`), where `q1` and `q3` are the first and third
    quartiles, respectively, and `iqr` is the interquartile range (`q3` - `q1`).
    """

    _iqr_multiplier: float

    def __init__(self: "CalcBoundsIQR", iqr_multiplier: float = 1.5) -> None:
        """Initialize functor.

        Parameters
        ----------
        iqr_multiplier : float, default 1.5
            The multiplier applied to the IQR to determine the bounds.
        """
        self._iqr_multiplier = iqr_multiplier

    def __call__(
        self: "CalcBoundsIQR", data: typing.Sequence[float]
    ) -> typing.Tuple[float, float]:
        """Calculate the lower and upper bounds of the input data using the IQR
        method.

        Parameters
        ----------
        data : np.ndarray
            The input data for which the bounds are calculated.

        Returns
        --------
        bounds : Tuple[float, float]
            Lower and upper calculated bounds.
        """
        if not any(data):
            return 0.0, 0.0
        quartile1, quartile3 = np.percentile(data, [25, 75])
        iqr = quartile3 - quartile1
        lower_bound = quartile1 - (self._iqr_multiplier * iqr)
        upper_bound = quartile3 + (self._iqr_multiplier * iqr)
        return lower_bound, upper_bound



import typing

import pandas as pd


def robust_groupby(
    data: pd.DataFrame,
    by: typing.List[typing.Optional[str]],
    **kwargs,
) -> typing.Iterable[
    typing.Tuple[typing.Union[typing.Tuple, str, None], pd.DataFrame]
]:
    """Perform a grouping operation on a pandas DataFrame that is robust to
    `by` sequences that are empty or contain None values.

    Extends the standard pandas groupby functionality by handling cases
    where grouping keys may contain None values.

    Parameters
    ----------
    data : pd.DataFrame
        The DataFrame to be grouped.
    by : typing.List[typing.Optional[str]]
        A list of column names to group by. None values in the list are ignored.
    **kwargs
        Additional keyword arguments to be passed to pandas.DataFrame.groupby.

    Returns
    -------
    typing.Iterable[typing.Tuple[typing.Union[typing.Tuple, str, None], pd.DataFrame]]
        An iterable of tuples, each containing contains the group key(s) and
        the corresponding grouped DataFrame.

        If 'by' is empty after filtering None, returns a tuple with
        None as the key and the original DataFrame.
    """
    by = list(filter(bool, by))
    if by:
        return data.groupby(by=by, **kwargs)
    else:
        return [(None, data)]



import itertools as it

from matplotlib import pyplot as plt
import pandas as pd
from sklearn.manifold import TSNE
from sklearn.datasets import load_iris
import seaborn as sns

import outset
from outset import patched as otst_patched


# adapted from https://www.datatechnotes.com/2020/11/tsne-visualization-example-in-python.html
def test_cluster_regplot():
    iris = load_iris()
    x = iris.data
    y = iris.target
    tsne = TSNE(n_components=2, verbose=1, random_state=123)
    z = tsne.fit_transform(x)

    df = pd.DataFrame()
    df["y"] = y
    df["comp-1"] = -z[:, 0]
    df["comp-2"] = z[:, 1]

    palette = sns.color_palette("hls", 6)
    og = outset.OutsetGrid(
        data=df,
        x="comp-1",
        y="comp-2",
        col="y",
        hue="y",
        col_wrap=2,
        palette=palette,
        marqueeplot_source_kws={
            "leader_stretch": 0.07,
            "mark_retract": 0.25,
        },
        zorder=4,
    )

    og.map_dataframe(
        otst_patched.regplot,
        x="comp-1",
        y="comp-2",
        palette=palette,
    )

    og.broadcast(sns.despine)
    og.marqueeplot()

    outpath = "/tmp/test_regplot.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")



import typing

from matplotlib import axes as mpl_axes
import numpy as np

from .calc_aspect_ import calc_aspect
from .set_aspect_ import set_aspect


def equalize_aspect(axs: typing.List[mpl_axes.Axes]) -> float:
    """Equalize the aspect ratio across multiple matplotlib Axes objects.

    This function calculates the geometric mean of the aspect ratios of a list
    of Axes objects and sets this common aspect ratio to all Axes in the list.
    The aspect ratio is defined as the ratio of the range of the y-axis to the
    range of the x-axis.

    Parameters
    ----------
    axs : typing.List[mpl_axes.Axes]
        A list of matplotlib Axes objects whose aspect ratios are to be
        equalized.

    Returns
    -------
    float
        The new common aspect ratio set for all Axes objects.
    """
    if not axs:
        return 1.0
    aspects = [calc_aspect(a) for a in axs]
    new_aspect = np.sqrt(min(aspects) * max(aspects))  # geometric mean
    for ax in axs:
        set_aspect(ax, new_aspect)

    assert all(np.isclose(calc_aspect(a), new_aspect) for a in axs)
    return new_aspect



import matplotlib.pyplot as plt

from outset import mark as otst_mark


def test_MarkAlphabeticalBadges():
    ax = plt.gca()  # Get the current axes

    ftor = otst_mark.MarkAlphabeticalBadges()
    # Call the function with test values
    ftor(x=0.5, y=0.5, ax=ax, color="red")
    ftor(
        x=1.5,
        y=1.5,
        ax=ax,
        color="red",
        scale_letter=0.6,
    )

    # Save and print the path of the output image
    outpath = "/tmp/test_MarkAlphabeticalBadges.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")



import matplotlib.pyplot as plt
import pytest
import seaborn as sns

import outset as otst


@pytest.mark.integration
def test_make_example_hueless():
    og = otst.OutsetGrid(
        data=sns.load_dataset("iris").dropna(),
        x="petal_width",
        y="petal_length",
        col="species",
        col_wrap=2,
        color=sns.color_palette()[1],
        marqueeplot_kws={
            "mark_glyph": otst.mark.MarkAlphabeticalBadges,
        },
        marqueeplot_source_kws={"leader_tweak": otst.tweak.TweakReflect()},
        marqueeplot_outset_kws={
            "leader_tweak": otst.tweak.TweakReflect(vertical=True)
        },
    )
    og.map_dataframe(
        sns.kdeplot, x="petal_width", y="petal_length", legend=False, zorder=0
    )
    og.map_dataframe_outset(
        sns.scatterplot,
        x="petal_width",
        y="petal_length",
        legend=False,
        zorder=0,
    )
    og.marqueeplot()

    outpath = "/tmp/test_make_example_hueless.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")


@pytest.mark.integration
def test_make_example_huefull():
    og = otst.OutsetGrid(
        data=sns.load_dataset("penguins").dropna(),
        x="bill_length_mm",
        y="bill_depth_mm",
        col="island",
        hue="species",
        marqueeplot_source_kws={
            "leader_tweak": otst.tweak.TweakSpreadArea(
                spread_factor=6, xlim=(47.5, 52)
            ),
        },
    )
    og.map_dataframe(
        sns.scatterplot, x="bill_length_mm", y="bill_depth_mm", legend=False
    )
    og.marqueeplot()
    og.set_axis_labels("bill length (mm)", "bill depth (mm)")
    og.add_legend()

    outpath = "/tmp/test_make_example_huefull.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")


@pytest.mark.integration
def test_make_example_sourceless():
    og = otst.OutsetGrid(
        data=sns.load_dataset("penguins").dropna(),
        x="bill_length_mm",
        y="bill_depth_mm",
        col="island",
        hue="species",
        include_sourceplot=False,
        marqueeplot_source_kws={
            "leader_tweak": otst.tweak.TweakSpreadArea(
                spread_factor=6, xlim=(47.5, 52)
            ),
        },
    )
    og.map_dataframe(
        sns.scatterplot, x="bill_length_mm", y="bill_depth_mm", legend=False
    )
    og.marqueeplot()
    og.add_legend()

    outpath = "/tmp/test_make_example_sourceless.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")


@pytest.mark.integration
def test_make_example_singleton():
    og = otst.OutsetGrid(
        data=[(73.5, 23.5, 78.5, 31.5)],
        color=sns.color_palette()[-1],
        marqueeplot_kws={
            "mark_glyph": otst.mark.MarkAlphabeticalBadges,
            "frame_outer_pad": 0.2,
            "frame_outer_pad_unit": "inches",
            "frame_face_kws": {"facecolor": "none"},
        },
    )
    og.broadcast(
        sns.scatterplot,
        data=sns.load_dataset("mpg").dropna(),
        x="horsepower",
        y="mpg",
        hue="origin",
        size="weight",
        sizes=(40, 400),
        alpha=0.5,
        palette="muted",
        legend=False,
        zorder=0,
    )
    og.marqueeplot(equalize_aspect=False)

    outpath = "/tmp/test_make_example_singleton.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")



import typing

import numpy as np
import pytest

from outset.tweak import TweakSpreadArea


def test_tweak_spread_area_init():
    # Test initialization with default parameters
    tsa_default = TweakSpreadArea()
    assert tsa_default._spread_factor == (2.0, 2.0)
    assert tsa_default._xlim is None
    assert tsa_default._ylim is None

    # Test initialization with custom parameters
    tsa_custom = TweakSpreadArea(
        spread_factor=(1.5, 2.5), xlim=(0, 10), ylim=(0, 5)
    )
    assert tsa_custom._spread_factor == (1.5, 2.5)
    assert tsa_custom._xlim == (0, 10)
    assert tsa_custom._ylim == (0, 5)


@pytest.mark.parametrize(
    "outer_vertex_and_result",
    zip([(-1, 0), (5, 2), (9, 4)], [(-1, -2.5), (5, 1.5), (13, 5.5)]),
)
def test_tweak_spread_area_call(
    outer_vertex_and_result: typing.Tuple[typing.Tuple[float, float]]
):
    tsa = TweakSpreadArea(spread_factor=(2, 2), xlim=(0, 10), ylim=(0, 5))
    leader_vertices = [(-1, 0), (4, 7), (10, 15), outer_vertex_and_result[0]]
    expected_result = [(-1, 0), (4, 7), (10, 15), outer_vertex_and_result[1]]

    modified_vertices = tsa(leader_vertices)
    assert np.allclose(modified_vertices, expected_result)
